/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal.storage.eclipse;

import static org.eclipse.core.resources.IResourceDelta.ADDED;
import static org.eclipse.core.resources.IResourceDelta.CHANGED;
import static org.eclipse.core.resources.IResourceDelta.REMOVED;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getBundle;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getConfigurationElements;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.loadClass;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IEclipseResourceStorageLocationProvider;
import org.fortiss.tooling.kernel.extension.IStorageProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.extension.data.ModelStorageError;
import org.fortiss.tooling.kernel.service.IEclipseResourceStorageService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.osgi.framework.Bundle;

/**
 * This class implements the persistency service behavior for Eclipse file
 * storage.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: CD1228D699275835E9DA2E9079A5F264
 */
public class EclipseResourceStorageProvider implements IEclipseResourceStorageService,
		IResourceChangeListener, IResourceDeltaVisitor, IStorageProvider {

	/** The prototype provider extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.eclipseResourceStorageLocationProvider";

	/** The prototype provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "storageLocationProvider";

	/** Stores the registered storage providers. */
	private final List<IEclipseResourceStorageLocationProvider> storageProviderList =
			new ArrayList<IEclipseResourceStorageLocationProvider>();

	/** Cache of models loaded so far. */
	private final Map<IFile, ModelContext> loadedContexts = new HashMap<IFile, ModelContext>();

	/** List of failed model storage files. */
	private final List<ModelStorageError> errorFiles = new ArrayList<ModelStorageError>();

	/** Map from toplevel elements to context. */
	private final Map<EObject, ModelContext> rootElementContexts =
			new IdentityHashMap<EObject, ModelContext>();

	/** Constructor. */
	public EclipseResourceStorageProvider() {
		setupLocationProviders();
		searchWorkspaceForModels();
		// register resource change listener last, since on-disk changes
		// are discovered during workspace search, thus introducing a possible
		// cyclic call during provider initialization
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
	}

	/** Searches existing models in Eclipse project roots. */
	private synchronized void searchWorkspaceForModels() {
		for(IProject project : ResourcesPlugin.getWorkspace().getRoot().getProjects()) {
			if(!project.isOpen()) {
				continue;
			}
			searchResourceForModels(project);
		}
	}

	/** Searches the given container for models. */
	private void searchResourceForModels(IContainer container) {
		try {
			for(IResource res : container.members()) {
				if(res instanceof IFile) {
					IFile file = (IFile)res;
					if(checkLocationProvider(file)) {
						loadContext(file);
					}
				} else if(res instanceof IFolder) {
					searchResourceForModels((IFolder)res);
				}
			}
		} catch(CoreException e) {
			error(ToolingKernelActivator.getDefault(),
					"Core Exception while scanning Eclipse workspace", e);
		}
	}

	/**
	 * Searches all location providers and loads the model context if some
	 * location provider validates the given file as model storage file.
	 */
	private boolean checkLocationProvider(IFile file) {
		if(file.getFileExtension() == null) {
			return false;
		}
		for(IEclipseResourceStorageLocationProvider provider : storageProviderList) {
			if(provider.isStorageLocation(file)) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public void resourceChanged(IResourceChangeEvent event) {
		synchronized(this) {
			if(event.getDelta() == null) {
				return;
			}
			try {
				event.getDelta().accept(this);
			} catch(final CoreException e) {
				error(ToolingKernelActivator.getDefault(), "Kernel Core Ecxeption!", e);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean visit(IResourceDelta delta) {
		if(delta.getResource() instanceof IFile &&
				checkLocationProvider((IFile)delta.getResource())) {
			runWorkspaceChangeJob((IFile)delta.getResource(), delta.getKind());
		}
		return true;
	}

	/** Runs resource change job as workspace job. */
	private void runWorkspaceChangeJob(final IFile file, final int changeKind) {
		new WorkspaceJob("ModelFileChangedJob") {
			@Override
			public IStatus runInWorkspace(IProgressMonitor monitor) {
				synchronized(EclipseResourceStorageProvider.this) {
					if(changeKind == ADDED) {
						loadContext(file);
					} else if(changeKind == REMOVED && isLoaded(file)) {
						unloadContext(file);
					} else if(changeKind == CHANGED) {
						if(isLoaded(file)) {
							handleChange(file);
						} else {
							loadContext(file);
						}
					}
					// Refresh the top-level elements
					IPersistencyService.INSTANCE
							.refreshTopLevelElements(EclipseResourceStorageProvider.this);
				}
				return Status.OK_STATUS;
			}
		}.schedule();
	}

	/** Handles a change of the given file (which is actually managed). */
	private void handleChange(IFile file) {
		ModelContext context = loadedContexts.get(file);
		// this avoids to reload the context when itself caused the disk change
		if(!context.getAndResetLastChangeWasCausedByModelContext()) {
			unloadContext(file);
			loadContext(file);
		}
	}

	/** Returns whether the given file has been loaded by this manager. */
	public synchronized boolean isLoaded(IFile file) {
		return loadedContexts.containsKey(file);
	}

	/** Loads a file and creates a new context from its contents. */
	private void loadContext(IFile file) {
		try {
			ModelContext mc = new ModelContext(file);
			loadedContexts.put(file, mc);
			rootElementContexts.put(mc.getRootModelElement(), mc);
		} catch(Exception ioex) {
			String msg = "Exception while loading model file: " + file.getName();
			errorFiles.add(new FileModelStorageError(msg, ioex, file));
			error(ToolingKernelActivator.getDefault(), msg, ioex);
		}
	}

	/** Unloads the context and destroys it. */
	private void unloadContext(IFile file) {
		ModelContext context = loadedContexts.remove(file);
		if(context != null) {
			rootElementContexts.remove(context.getRootModelElement());
			context.destroy();
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<ITopLevelElement> getTopLevelElements() {
		List<ITopLevelElement> result = new ArrayList<ITopLevelElement>();
		result.addAll(rootElementContexts.values());
		return result;
	}

	/** Initializes the storage provider list from plugin extensions. */
	private void setupLocationProviders() {
		for(IConfigurationElement ce : getConfigurationElements(EXTENSION_POINT_NAME,
				CONFIGURATION_ELEMENT_NAME)) {
			Bundle bundle = getBundle(ce);
			try {
				Class<?> handlerClass = loadClass(ce.getAttribute("provider"), bundle);
				IEclipseResourceStorageLocationProvider provider =
						(IEclipseResourceStorageLocationProvider)handlerClass.getConstructor()
								.newInstance();
				storageProviderList.add(provider);
			} catch(Exception ex) {
				error(ToolingKernelActivator.getDefault(), ex.getMessage(), ex);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<IFile> getModelElementResources() {
		return new ArrayList<IFile>(loadedContexts.keySet());
	}

	/** {@inheritDoc} */
	@Override
	public List<ModelStorageError> getStorageErrors() {
		checkFiles();
		return errorFiles;
	}

	/** Checks the erroneous files for existence. */
	private void checkFiles() {
		List<ModelStorageError> iter = new ArrayList<ModelStorageError>(errorFiles);
		for(ModelStorageError err : iter) {
			if(!((FileModelStorageError)err).file.exists()) {
				errorFiles.remove(err);
			}
		}
	}

	/** Class for file-based model errors. */
	private static class FileModelStorageError extends ModelStorageError {
		/** The file. */
		public IFile file;

		/** Constructor. */
		public FileModelStorageError(String msg, Exception cause, IFile file) {
			super(msg, cause);
			this.file = file;
		}
	}
}
